"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var reduce_css_calc_1 = __importDefault(require("reduce-css-calc"));
var css_1 = __importDefault(require("css"));
/**
 * Checks whether a node parsed by @reworckss/css is a css rule
 *
 * @param {css.Node} node
 * @returns Whether the node can be a css.Rule
 */
function isRule(node) {
    // could still be Page AtRule 
    return node && node.type === 'rule';
}
/**
 * Checks whether a node parsed by @reworckss/css is a css declaration
 *
 * @param {css.Node} node
 * @returns Whether the node can be a css.Declaration
 */
function isDeclaration(node) {
    return node && node.type === 'declaration';
}
/**
 * Checks whether a declaration parsed by @reworckss/css is a css variable
 *
 * @param {css.Declaration} declaration
 * @returns Whether the declaration is a variable
 */
function isVariable(declaration) {
    var _a;
    return (_a = declaration.property) === null || _a === void 0 ? void 0 : _a.startsWith('--');
}
/**
 * Replaces variable reads within the string with static values if found within the provided lookup
 *
 * @param {VariableDict} lookup
 * @param {string} value The value to carry out replacement on
 * @param {string[]} dependencies The variables that are read in this value
 * @returns The value with its dependencies replaced
 */
function replaceVariables(lookup, value, dependencies) {
    var res = value;
    for (var _i = 0, dependencies_1 = dependencies; _i < dependencies_1.length; _i++) {
        var dependency = dependencies_1[_i];
        var dependencyValue = lookup[dependency];
        if (dependencyValue) {
            // maybe should add a step to remove spaces at the beginning
            // instead of trying to handle this at this step
            var regex = new RegExp('var\\(\\s*' + dependency + '\\s*\\)', 'g');
            res = res.replace(regex, dependencyValue);
        }
    }
    return res;
}
/**
 * Finds all css variables read in a string
 *
 * @param {string} value
 * @returns An array of css variables that were read within this string
 */
function variablesInValue(value) {
    return value.match(/(?<=var\()(--.+?)(?=\))/g) || [];
}
/**
 * Iterates over a stylesheet's nodes to find all variables within the provided selector's scope
 *
 * @param {css.Stylesheet} stylesheet
 * @param {string} [selector] Css selector to look for variables in. Has to be exact
 * @returns A map of variables to their values
 */
function getVariablesFromStylesheet(stylesheet, selector) {
    var res = {};
    for (var _i = 0, _a = stylesheet.stylesheet.rules; _i < _a.length; _i++) {
        var node = _a[_i];
        if (isRule(node) && (!selector || node.selectors && node.selectors.length === 1 && node.selectors.includes(selector))) {
            for (var _b = 0, _c = (node.declarations || []); _b < _c.length; _b++) {
                var declaration = _c[_b];
                if (isDeclaration(declaration) &&
                    declaration.property &&
                    isVariable(declaration)) {
                    var property = declaration.property, value = declaration.value;
                    if (property && value)
                        res[property] = value;
                }
            }
        }
    }
    return res;
}
/**
 * Looks for variables that are read within the VariableDict's values but do not exist within the VariableDict
 *
 * @param {VariableDict} variables
 * @returns Variables that don't exist as a VariableDict<true> for easier lookup
 */
function getUnresolvedNames(variables) {
    var res = {};
    for (var variable in variables) {
        var value = variables[variable];
        var dependencies = variablesInValue(value);
        for (var _i = 0, dependencies_2 = dependencies; _i < dependencies_2.length; _i++) {
            var dependency = dependencies_2[_i];
            if (!variables[dependency])
                res[dependency] = true;
        }
    }
    return res;
}
/**
 * Parses strings as css stylesheets to create a reference for their absolute/resolved string values
 *
 * @param {string[]} content List of strings, in order, to parse for variables
 * @param {string} [selector=':root'] A selector to scope the search for variables
 * @returns {{resolved: VariableDict, failed: string[]}} Resolved variables with absolute values + a list of the variables that failed to be resolved.
 */
module.exports = function resolveVariables(content, selector) {
    if (selector === void 0) { selector = ':root'; }
    var rawVariables = content.reduce(function (previous, contents) { return (__assign(__assign({}, previous), getVariablesFromStylesheet(css_1.default.parse(contents), selector))); }, {});
    var copy = function (value) { return JSON.parse(JSON.stringify(value)); };
    var resolved = {};
    /**
     * Some inefficiency in iteration here - see the unresolved/nextUnresolved interactions w the loops
     */
    var failed = getUnresolvedNames(rawVariables);
    var unresolved = copy(rawVariables);
    var nextUnresolved = {};
    while (Object.keys(unresolved).length) {
        for (var variable in unresolved) {
            if (failed[variable]) {
                continue;
            }
            var value = unresolved[variable];
            if (!value) {
                failed[variable] = true;
            }
            else {
                var dependsOnVariables = variablesInValue(value);
                if (dependsOnVariables === null || dependsOnVariables === void 0 ? void 0 : dependsOnVariables.length) {
                    if (dependsOnVariables.some(function (variable) { return failed[variable]; })) {
                        failed[variable] = true;
                        continue;
                    }
                    nextUnresolved[variable] = replaceVariables(resolved, value, dependsOnVariables);
                }
                else {
                    resolved[variable] = reduce_css_calc_1.default(value);
                }
            }
        }
        unresolved = copy(nextUnresolved);
        nextUnresolved = {};
    }
    return {
        resolved: resolved,
        failed: Object.keys(failed)
    };
};
